%% OLS ESTIMATES OF A VAR(p) MODEL, conditioned on the first p observations
% MODEL: y_t=v+A1*y_t-1+...+Ap*y_t-p+u_t, 
% where y_t is a kx1 vector, each Ai a KxK matrix (i=1,...,p) and u_t a Kx1
% vector.
% INPUTS: %%
% y - a KxT matrix of data, K num. of var., T num. of obs.
% p - lag length of VAR model
% OUTPUTS: %%
% K - number of eq.
% T - number of obs.
% p - selected lag length
% beta - a Kx(1+Kp) matrix of estimated coeff., beta=[v A1 ... Ap], where v
% is a Kx1 vector of constants, Ai a KxK matrix of i-th lag estimated coeff
% and i=1,...,p
% cvm - a KxK var-cov. matrix of est. resid.
% sbeta - a Kx(1+Kp) matrix of std. err. of the est. coeff.
% tbeta - a Kx(1+Kp) matrix of t stat. of the est. coeff
% resids - est. residuals

function [VARcoefficients, residuals, cvm, SSR] = CalculateVarStatistics(data, lags)


%% PRELIMINARIES %%
% get dimensions
[K, T]=size(data);

%% TRANSFORMATION of THE MODEL %%
% Into the following form: Y=beta*Z+U, where:
% Y:=[y_p+1 ... y_T] is a Kx(T-p) matrix, we condition on the first p obs.
% beta:= [v A1 ... Ap], as before
% Z:= [Z_p ... Z_T-1] is a (1+pK)x(T-p+1) matrix, where Z_t=[1 y_t ...
% y_t-p+1]', y_t as before, t=p,...,T-1 (see Lutkepohl, 2007, p.70)
% U:=[u_p+1 ... u_T] is a Kx(T-p) matrix of est. resid.


%% Builds the Y matrix
Y = data(:,lags+1:T);

%% Builds the Z matrix
Z=zeros(1+K*lags,T-lags);
Z(1,:)=1;
for i=0:lags-1
    Z(i*K+2:(i+1)*K+1,:) = data(:,lags-i:T-i-1);
end

%% ESTIMATES %%
% creates beta mat. (see Lutkepohl, 2007, p.72, formula 3.2.10)
B=(Y*Z')/(Z*Z'); 
VARcoefficients = reshape(B, K + lags*K^2, 1);
%% creates residuals
residuals = Y - B*Z;
% creates covar mat. of resid. (See Lutkepohl, 2007, p.75, formula 3.2.18 and 3.2.19)
cvm=residuals*residuals'*1/(T - lags);

cov_beta = kron((Z*Z')^-1, cvm);
SSR = residuals*residuals';














