% This is the class that describes the VAR model where structural
% parameters are estimated by non-lineral optimization
classdef classHVAR
    properties
        lags;
        startingValue;
        Y
        T
        Z
        residuals0
        B0
        Sigma0
        Lambda0
        VARcoefficients0
        regimeRealisations
        optimisationOptions
        logL
        VARcoefficients
        B
        Lambda
        Sigma
        shortRunRestrictions
        longRunRestrictions
        shortRunRestrictionsMatrix
        longRunRestrictionsMatrix
        stdErrors
        standardErrorsB
        standardErrorsLambda
        residuals
        nonLinconstr
        waldTest
        waldTestNK
        LMtest
    end
    methods
       %% constructor
        function obj = classHVAR(shortRunRestrictions, longRunRestrictions, shortRunRestrictionsMatrix, longRunRestrictionsMatrix, Y, lags, S, sv)
            obj.Y = Y;
            obj.lags = lags;
            obj.regimeRealisations = S;
            obj.startingValue = sv;
            obj.T = size(obj.Y);
            obj.Z = CalculateMatrLaggedVariables(obj.Y, obj.T, obj.lags);
            [obj.VARcoefficients0, obj.residuals0, ~, ~] = CalculateVarStatistics(obj.Y', obj.lags);
            obj.residuals0 = obj.residuals0';
            obj.B0 =  0.5*(1/obj.T(1,1)*(obj.residuals0'*obj.residuals0))^0.5 + 1/100000*randn(obj.T(1,2),obj.T(1,2));
            startOfRegime = 1;
            endOfRegime = 0;
            for s = 1: max(obj.regimeRealisations)
                obj.Lambda0((s-1)*obj.T(1,2) + 1 : s*obj.T(1,2), :) = obj.startingValue *eye(obj.T(1,2));
                endOfRegime = endOfRegime + sum(obj.regimeRealisations == s);
                obj.Sigma0((s-1)*obj.T(1,2) + 1 : s*obj.T(1,2), :) = 1/(endOfRegime - startOfRegime)*(obj.residuals0(startOfRegime : endOfRegime, :)'*obj.residuals0(startOfRegime : endOfRegime, :));
                startOfRegime = endOfRegime + 1;
            end
            obj.shortRunRestrictions = shortRunRestrictions;
            obj.longRunRestrictions = longRunRestrictions;
            obj.shortRunRestrictionsMatrix = shortRunRestrictionsMatrix;
            obj.longRunRestrictionsMatrix = longRunRestrictionsMatrix;
            obj.optimisationOptions = optimset('Algorithm', 'interior-point', 'Display', 'off', 'MaxFunEvals', 20000, ...
                'MaxIter', 20000, 'Hessian','bfgs', 'DerivativeCheck','on','Diagnostics','off','GradObj','off','LargeScale','off');
        end
        
       %% optimization method
        function obj = EstimateHVAR(obj)
            disp('Estimating...')
            structuralParameters = obj.B0;
            for s = 2 : max(obj.regimeRealisations)
                structuralParameters(:, obj.T(1,2) + s-1 ) = diag(GetSigmaForRegime(obj.Lambda0, obj.T, s));
            end
            
            VARcoefficients = GLS(obj.Z, obj.Y, obj.T, obj.lags, obj.Sigma0, obj.regimeRealisations);
            lowerBound = GetLowerBound(obj.T, max(obj.regimeRealisations));
            
            selectionMatrix = zeros(obj.T(1,2)-1, obj.T(1,2));
            for k = 1 : (obj.T(1,2)-1)
                selectionMatrix(k, k:k+1) = [1 -1];
            end
            
            convergenceCriteria(1, 1) = 1;
            convergenceCriteria(1, 2) = 1;
            roundOfIteration = 1;
            
            while convergenceCriteria(roundOfIteration,2) > 0.000000001
                
                matrixOfLongRunEffects = eye(obj.T(1,2), obj.T(1,2));
                for w = 1 : obj.lags
                    matrixOfLongRunEffects = matrixOfLongRunEffects - GetCoefficient(VARcoefficients, obj.T, w);
                end
                
                if obj.shortRunRestrictions == 1 && obj.longRunRestrictions == 1
                    ceq = @(structuralParameters) [
                        obj.shortRunRestrictionsMatrix * reshape(structuralParameters( 1 :obj.T(1,2), 1:   obj.T(1,2)), obj.T(1,2)^2, 1)
                        obj.longRunRestrictionsMatrix * kron( eye(obj.T(1,2)), matrixOfLongRunEffects^-1 ) * reshape(structuralParameters( 1 :obj.T(1,2), 1:   obj.T(1,2)), obj.T(1,2)^2, 1)
                        ];
                elseif obj.shortRunRestrictions == 1 && obj.longRunRestrictions ~= 1
                    ceq = @(structuralParameters) [
                        obj.shortRunRestrictionsMatrix * reshape(structuralParameters( 1 :obj.T(1,2), 1:   obj.T(1,2)), obj.T(1,2)^2, 1)
                        ];
                elseif obj.shortRunRestrictions ~= 1 && obj.longRunRestrictions == 1
                    ceq = @(structuralParameters) [
                        obj.longRunRestrictionsMatrix * kron( eye(obj.T(1,2)), matrixOfLongRunEffects^-1 ) * reshape(structuralParameters( 1 :obj.T(1,2), 1:   obj.T(1,2)), obj.T(1,2)^2, 1)
                        ];
                else
                    ceq = @(structuralParameters)[];
                end
                
                if obj.shortRunRestrictions == 1 || obj.longRunRestrictions == 1
                    obj.nonLinconstr = @(structuralParameters)deal( [] ,ceq(structuralParameters));
                else
                    obj.nonLinconstr = @(structuralParameters)deal([], []);
                end
                
                % estimated B and L
                [structuralParameters, minV] = fmincon( @(structuralParameters) MinimizeLikelihood(obj.Y, obj.T, obj.lags, structuralParameters, obj.regimeRealisations, obj.Z, VARcoefficients), structuralParameters , [], [] , [] , [] , lowerBound , [] , obj.nonLinconstr, obj.optimisationOptions);
                B = structuralParameters(1:obj.T(1,2), 1:1:obj.T(1,2));
                Lambda = eye(obj.T(1,2));
                for s = 2 : max(obj.regimeRealisations)
                    Lambda = [Lambda; diag(structuralParameters(:,obj.T(1,2) + s - 1))];
                end
                
                if obj.shortRunRestrictions ~= 1
                    currentLambda = diag(structuralParameters(:, obj.T(1,2)+1));
                    for i = 1: obj.T(1,2)-1
                        for j = i+1:obj.T(1,2)
                            if currentLambda(i,i) < currentLambda(j,j)
                                for s = 2 : max(obj.regimeRealisations)
                                    L_temp = currentLambda(i,i);
                                    currentLambda(i,i) = currentLambda(j,j);
                                    currentLambda(j,j) = L_temp;
                                    Lambda((s-1)*obj.T(1,2) + 1 : s*obj.T(1,2), :) = currentLambda;
                                    if s < max(obj.regimeRealisations)
                                        currentLambda = diag(structuralParameters(:, obj.T(1,2)+s));
                                    end
                                end

                                B_temp(:,1) =  B(:,i);
                                B(:,i) = B(:,j);
                                B(:,j) = B_temp(:, 1);

                            end
                        end
                    end
                end
                
                for s = 1:max(obj.regimeRealisations)
                    Sigma((s-1)*obj.T(1,2) + 1 : s*obj.T(1,2), :)  = B * GetSigmaForRegime(Lambda, obj.T, s) * B';
                end
                
                VARcoefficients = GLS(obj.Z, obj.Y, obj.T, obj.lags, Sigma, obj.regimeRealisations);
                convergenceCriteria(roundOfIteration+1,1) = minV;
                convergenceCriteria(roundOfIteration+1,2) = abs( ( convergenceCriteria(roundOfIteration+1,1) - convergenceCriteria(roundOfIteration,1) ) / convergenceCriteria(roundOfIteration,1) );
                roundOfIteration = roundOfIteration + 1;
            end
            
            %  structural parameters and VAR parameters
            obj.VARcoefficients = VARcoefficients;
            obj.logL = -1*minV;
            obj.B = B;
            obj.Lambda = Lambda;
            obj.Sigma = Sigma;
            obj.residuals = GetResiduals(obj.T, obj.Y, obj.Z, obj.lags, obj.VARcoefficients);
        end
        
       %% calculation of test statistic
        function obj = test(obj)
            test_count = 1;
            for s = 0: obj.T(1,2)-1
                for r = 2:obj.T(1,2)-s
                    z_m = zeros(obj.T(1,2), max(obj.regimeRealisations));
                    w_m = zeros(obj.T(1,2), max(obj.regimeRealisations));
                    kappa = zeros ( max(obj.regimeRealisations), 1);
                    for regime = 1 : max(obj.regimeRealisations)
                        for k = 1: obj.T(1,2)
                            z_m(k, regime) = sum((obj.residuals(obj.regimeRealisations(:, 1) == regime, k) - mean(obj.residuals(obj.regimeRealisations(:, 1)== regime, k))) .^4);
                            sigma = GetSigmaForRegime(obj.Sigma, obj.T, regime); 
                            z_m(k, regime) = z_m(k, regime) - 6 * sigma(k,k)^2;
                            z_m(k, regime) = z_m(k, regime)/(sum(obj.regimeRealisations(:, 1) == regime) - 4);

                            w_m_t = (sum(obj.regimeRealisations(:, 1) == regime))/(sum(obj.regimeRealisations(:, 1) == regime) - 1);
                            w_m(k, regime) = w_m_t*(sigma(k,k)^2 - z_m(k, regime)/sum(obj.regimeRealisations(:, 1) == regime));
                        end
                        kappa(regime, 1) = 1/(3 * obj.T(1,2)) * sum(z_m(:, regime) ./ w_m(:, regime) ) - 1;
                    end
                    tau = sum(obj.regimeRealisations(:, 1) == 1) / (obj.T(1,1) - obj.lags);
                    c_tau2 = ( (1 + kappa(1,1)) / tau + ( (1 + kappa(2,1)) /(1-tau)))^-1;
                    c_tau2_nk = ( (1) / tau + ( (1) /(1-tau)))^-1;
                    lambda_sum = 0;
                    lambda2 = GetSigmaForRegime(obj.Lambda, obj.T, 2);
                    lambda_sum_r = 0;
                    for n = s+1 : s + r
                        lambda_sum = lambda_sum + log(lambda2(n, n));
                        lambda_sum_r = lambda_sum_r + lambda2(n,n);
                    end
                    lambda_sum_r = log(lambda_sum_r / r);
                    Qr = c_tau2 * (-1*(obj.T(1,1)-obj.lags) * lambda_sum + (obj.T(1,1)-obj.lags) * r * lambda_sum_r );
                    Qr_nk = c_tau2_nk * (-1*(obj.T(1,1)-obj.lags) * lambda_sum + (obj.T(1,1)-obj.lags) * r * lambda_sum_r );
                    degrees_of_freedom = 0.5 * (r+2) * (r-1);

                    Qr_star = 2 * Qr / ((r+2)*(r-1));
                    p_value_star = 1 - fcdf(Qr_star, 0.5*(r+2)*(r-1), (obj.T(1,1)-obj.lags - obj.T(1,2)*obj.lags - 1));
                    
                    p_value = 1 - chi2cdf(Qr, degrees_of_freedom);
                    p_value_nk = 1 - chi2cdf(Qr_nk, degrees_of_freedom);
                    
                    obj.waldTest(test_count, 1) = s;
                    obj.waldTest(test_count, 2) = r;
                    obj.waldTest(test_count, 3) = Qr;
                    obj.waldTest(test_count, 4) = degrees_of_freedom;
                    obj.waldTest(test_count, 5) = p_value;

                    obj.waldTest(test_count, 6) = Qr_star;
                    obj.waldTest(test_count, 7) = p_value_star;
                    
                    
                    obj.waldTestNK(test_count, 1) = s;
                    obj.waldTestNK(test_count, 2) = r;
                    obj.waldTestNK(test_count, 3) = Qr_nk;
                    obj.waldTestNK(test_count, 4) = degrees_of_freedom;
                    obj.waldTestNK(test_count, 5) = p_value_nk;                    
                    test_count = test_count + 1;
                end
            end
        end
        
       %% method to calculate strandard errors
        function obj = CalculateStandardErrors(obj)
            param = reshape(obj.B, obj.T(1,2)^2, 1);
            for s = 2 :max(obj.regimeRealisations)
                param = [param; diag(GetSigmaForRegime(obj.Lambda, obj.T, s))];
            end
            param = [param; obj.VARcoefficients];
            myDelta=1e-4*abs(param)+0.00000001;
            
            % outer product calculation as in Hamilton page 143
            approximatedDerivatives = zeros(obj.T(1,1)-obj.lags,numel(param));
            
            % calculate the initial likelihood value for each t.
            [~, logLikVec] = CalculateLogL(obj.T, obj.Y, obj.Z, obj.lags, obj.VARcoefficients, obj.Sigma, obj.regimeRealisations);
            
            for i = 1:numel(param)
                permutedParameters = param;
                permutedParameters(i) = param(i)+myDelta(i);
                
                permutedB = reshape(permutedParameters(1:obj.T(1,2)^2, 1), obj.T(1,2), obj.T(1,2)) ;
                
                permutedLambda = eye(obj.T(1,2));
                for s = 2 : max(obj.regimeRealisations)
                	permutedLambda((s-1)*obj.T(1,2) + 1 : s*obj.T(1,2), :) = diag(permutedParameters(obj.T(1,2)^2 + obj.T(1,2)*(s-2) + 1 : obj.T(1,2)^2 + obj.T(1,2)*(s-1) ));
                end
                
                for s = 1:max(obj.regimeRealisations)
                    permutedSigma((s-1)*obj.T(1,2) + 1 : s*obj.T(1,2), :)  = permutedB * GetSigmaForRegime(permutedLambda, obj.T, s) * permutedB';
                end
                
                permutedVARcoefficients = permutedParameters(obj.T(1,2)^2  + (max(obj.regimeRealisations) -1)*obj.T(1,2) + 1 : end);
                % update the likelihood values
                
                [~, logLikVecUpdate(:, i)] = CalculateLogL(obj.T, obj.Y, obj.Z, obj.lags, permutedVARcoefficients, permutedSigma, obj.regimeRealisations);
                
                % calculate the approximated derivative for each t and each parameter
                for t=1:obj.T(1,1) - obj.lags
                    approximatedDerivatives(t,i)=(logLikVecUpdate(t,i)-logLikVec(t,1))/(myDelta(i));
                end
            end
            
            scoreVector = sum(approximatedDerivatives, 1);
            % form the outer product matrix as in Hamilton  eq 5.8.4
            sum_s_Matrix=zeros(numel(param));
            for idx = 1 : obj.T(1,1)-obj.lags
                sum_s_Matrix = sum_s_Matrix + approximatedDerivatives(idx,:)' * approximatedDerivatives(idx,:);
            end
            outerProduct = sum_s_Matrix/(obj.T(1,1) - obj.lags );
            standardErrors(:, 1) = sqrt(diag(inv(outerProduct * (obj.T(1,1) - obj.lags ) )));
            obj.stdErrors = standardErrors(:, 1);
            obj.standardErrorsB = reshape(standardErrors(1:obj.T(1,2)^2, 1), obj.T(1,2), obj.T(1,2));
            for s = 2 : max(obj.regimeRealisations)
                obj.standardErrorsLambda((s-1)*obj.T(1,2) + 1 : s*obj.T(1,2), :)  = diag(standardErrors(obj.T(1,2)^2 + obj.T(1,2)*(s-2) + 1 : obj.T(1,2)^2 + obj.T(1,2)*(s-1), 1));
            end
            
            % Wald test for the structural parameters
            hessian = (outerProduct*(obj.T(1,1) - obj.lags))^-1;
            numberOfRestrictions = size(obj.shortRunRestrictionsMatrix, 1);
            
        end
       
    end
end